from fastapi import APIRouter, HTTPException
from llamphouse.core.database import database as db
from ..types.message import DeleteMessageResponse, CreateMessageRequest, MessageListResponse, Attachment, MessageObject, TextContent, ImageFileContent, ModifyMessageRequest
from typing import List, Optional


router = APIRouter()

@router.post("/threads/{thread_id}/messages", response_model=MessageObject)
async def create_message(thread_id: str, request: CreateMessageRequest):
    try:
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        message = db.insert_message(thread_id, request)
        
        return MessageObject(
                    id=message.id,
                    role=message.role,
                    content=[
                        TextContent(text=message.content) if isinstance(message.content, str) else ImageFileContent(image_file=message.content)
                        for message.content in [message.content]
                    ],
                    metadata=message.meta,
                    status=message.status,
                    incomplete_details=message.incomplete_details,
                    completed_at=message.completed_at,
                    incomplete_at=message.incomplete_at,
                    assistant_id=message.assistant_id,
                    run_id=message.run_id,
                    attachments=[
                        Attachment(file_id=attachment['file_id'], tool=attachment.get('tool')) for attachment in (message.attachments or [])
                    ],
                    created_at=int(message.created_at.timestamp()),
                    thread_id=message.thread_id
                ) 
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")

@router.get("/threads/{thread_id}/messages", response_model=MessageListResponse)
async def list_messages(thread_id: str, limit: int = 20, order: str = "desc", after: Optional[str] = None, before: Optional[str] = None):
    try:
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        messages = db.get_messages_by_thread_id(
            thread_id=thread_id,
            limit=limit + 1,
            order=order,
            after=after,
            before=before
        )
        has_more = len(messages) > limit
        first_id = messages[0].id if messages else None
        last_id = messages[-1].id if messages else None

        return MessageListResponse(
            object="list",
            data=[
                MessageObject(
                    id=msg.id,
                    role=msg.role,
                    content=[
                        TextContent(text=msg.content) if isinstance(msg.content, str) else ImageFileContent(image_file=msg.content)
                        for msg.content in [msg.content]
                    ],
                    metadata=msg.meta,
                    status=msg.status,
                    incomplete_details=msg.incomplete_details,
                    completed_at=msg.completed_at,
                    incomplete_at=msg.incomplete_at,
                    assistant_id=msg.assistant_id,
                    run_id=msg.run_id,
                    attachments=[
                        Attachment(file_id=attachment['file_id'], tool=attachment.get('tool')) for attachment in (msg.attachments or [])
                    ],
                    created_at=int(msg.created_at.timestamp()),
                    thread_id=msg.thread_id
                ) 
                for msg in messages
            ],
            first_id=first_id,
            last_id=last_id,
            has_more=has_more
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")

@router.get("/threads/{thread_id}/messages/{message_id}", response_model=MessageObject)
async def retrieve_message(thread_id: str, message_id: str):
    try:
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        message = db.get_message_by_id(message_id)
        if not message:
            raise HTTPException(status_code=404, detail="Message not found.")
        
        return  MessageObject(
                    id=message.id,
                    role=message.role,
                    content=[
                        TextContent(text=message.content) if isinstance(message.content, str) else ImageFileContent(image_file=message.content)
                        for message.content in [message.content]
                    ],
                    metadata=message.meta,
                    status=message.status,
                    incomplete_details=message.incomplete_details,
                    completed_at=message.completed_at,
                    incomplete_at=message.incomplete_at,
                    assistant_id=message.assistant_id,
                    run_id=message.run_id,
                    attachments=[
                        Attachment(file_id=attachment['file_id'], tool=attachment.get('tool')) for attachment in (message.attachments or [])
                    ],
                    created_at=int(message.created_at.timestamp()),
                    thread_id=message.thread_id
                ) 
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")

@router.post("/threads/{thread_id}/messages/{message_id}", response_model=MessageObject)
async def modify_message(thread_id: str, message_id: str, request: ModifyMessageRequest):
    try:
        message = db.update_message_metadata(thread_id, message_id, request.metadata)
        if not message:
            raise HTTPException(status_code=404, detail="Message not found.")
        
        return  MessageObject(
                    id=message.id,
                    role=message.role,
                    content=[
                        TextContent(text=message.content) if isinstance(message.content, str) else ImageFileContent(image_file=message.content)
                        for message.content in [message.content]
                    ],
                    metadata=message.meta,
                    status=message.status,
                    incomplete_details=message.incomplete_details,
                    completed_at=message.completed_at,
                    incomplete_at=message.incomplete_at,
                    assistant_id=message.assistant_id,
                    run_id=message.run_id,
                    attachments=[
                        Attachment(file_id=attachment['file_id'], tool=attachment.get('tool')) for attachment in (message.attachments or [])
                    ],
                    created_at=int(message.created_at.timestamp()),
                    thread_id=message.thread_id
                ) 
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")

@router.delete("/threads/{thread_id}/messages/{message_id}", response_model=DeleteMessageResponse)
async def delete_message(thread_id: str, message_id: str):
    try:
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        message = db.get_message_by_id(message_id)
        if not message:
            raise HTTPException(status_code=404, detail="Message not found.")

        deleted = db.delete_message(thread_id, message_id)
        if not deleted:
            return DeleteMessageResponse(
                id=thread_id,
                deleted=False
            )
        
        return DeleteMessageResponse(
            id=thread_id,
            deleted=True
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")



from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import StreamingResponse
from ..types.run import RunObject, RunCreateRequest, CreateThreadAndRunRequest, RunListResponse, ModifyRunRequest
from ..assistant import Assistant
from ..database import database as db
from typing import List, Optional
import time
import asyncio
import json

router = APIRouter()


@router.post("/threads/{thread_id}/runs", response_model=RunObject)
async def create_run(
    thread_id: str,
    request: RunCreateRequest,
    req: Request
) -> RunObject:
    try:
        assistants = req.app.state.assistants
        assistant = get_assistant_by_id(assistants, request.assistant_id)
        # store run in db
        run = db.insert_run(thread_id, run=request, assistant=assistant)

        # Check if the task exists
        task_key = f"{request.assistant_id}:{thread_id}"
        if task_key not in req.app.state.task_queues:
            print(f"Creating queue for task {task_key} in RUN")
            req.app.state.task_queues[task_key] = asyncio.Queue(maxsize=1)
            # raise HTTPException(status_code=404, detail="Task not found")

        # check if stream is enabled
        if request.stream:
            output_queue: asyncio.Queue = req.app.state.task_queues[task_key]

            # Streaming generator for SSE
            async def event_stream():
                while True:
                    try:
                        event = await asyncio.wait_for(output_queue.get(), timeout=10.0)  # Set timeout in seconds
                        if event is None:  # Stream completion signal
                            break
                        print(f"Event: {event['event']}")
                        # output_queue.task_done()
                        yield f"event: {event['event']}\ndata: {event['data']}\n\n"
                    except asyncio.TimeoutError:
                        print("TimeoutError: No event received within the timeout period")
                        yield f'''event: error\ndata: {json.dumps({
                            "error": "TimeoutError",
                            "message": "No event received within the timeout period"
                        })}\n\n'''
                        break

            # Return the streaming response
            return StreamingResponse(event_stream(), media_type="text/event-stream")
        
        return RunObject(
            id=run.id,
            created_at=time.mktime(run.created_at.timetuple()),
            thread_id=thread_id,
            assistant_id=assistant.id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")

@router.post("/threads/runs", response_model=RunObject)
async def create_thread_and_run(request: CreateThreadAndRunRequest, req: Request):
    try:
        thread = db.insert_thread(request.thread)

        for msg in request.thread.messages:
            if msg.role not in ["user", "assistant"]:
                raise HTTPException(status_code=400, detail="Invalid role. Must be 'user' or 'assistant'.")
            else:
                db.insert_message(thread_id=thread.id, message=msg)

        assistants = req.app.state.assistants
        assistant = get_assistant_by_id(assistants, request.assistant_id)
        # store run in db
        run = db.insert_run(thread.id, run=request, assistant=assistant)

        # Check if the task exists
        task_key = f"{request.assistant_id}:{thread.id}"
        if task_key not in req.app.state.task_queues:
            print(f"Creating queue for task {task_key} in RUN")
            req.app.state.task_queues[task_key] = asyncio.Queue(maxsize=1)
            # raise HTTPException(status_code=404, detail="Task not found")

        # check if stream is enabled
        if request.stream:
            output_queue: asyncio.Queue = req.app.state.task_queues[task_key]

            # Streaming generator for SSE
            async def event_stream():
                while True:
                    try:
                        event = await asyncio.wait_for(output_queue.get(), timeout=10.0)  # Set timeout in seconds
                        if event is None:  # Stream completion signal
                            break
                        print(f"Event: {event['event']}")
                        # output_queue.task_done()
                        yield f"event: {event['event']}\ndata: {event['data']}\n\n"
                    except asyncio.TimeoutError:
                        print("TimeoutError: No event received within the timeout period")
                        yield f'''event: error\ndata: {json.dumps({
                            "error": "TimeoutError",
                            "message": "No event received within the timeout period"
                        })}\n\n'''
                        break

            # Return the streaming response
            return StreamingResponse(event_stream(), media_type="text/event-stream")
        
        return RunObject(
            id=run.id,
            created_at=time.mktime(run.created_at.timetuple()),
            thread=thread.id,
            assistant_id=assistant.id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
  
def get_assistant_by_id(assistants: List[Assistant], assistant_id: str) -> Assistant:
    assistant = next((assistant for assistant in assistants if assistant.id == assistant_id), None)
    if not assistant:
        raise HTTPException(status_code=404, detail="Assistant not found")
    return assistant

@router.get("/threads/{thread_id}/runs", response_model=RunListResponse)
async def list_runs(thread_id: str, limit: int = 20, order: str = "desc", after: Optional[str] = None, before: Optional[str] = None) -> RunObject:
    try:
        runs = db.get_runs_by_thread_id(
            thread_id=thread_id,
            limit=limit + 1,
            order=order,
            after=after,
            before=before
        )
        has_more = len(runs) > limit
        first_id = runs[0].id if runs else None
        last_id = runs[-1].id if runs else None
        return  RunListResponse(
                    object="list",
                    data=[
                        RunObject(
                            id=run.id,
                            created_at=int(run.created_at.timestamp()),
                            thread_id=thread_id,
                            assistant_id=run.assistant_id,
                            status=run.status,
                            required_action=run.required_action,
                            last_error=run.last_error,
                            expires_at=run.expires_at,
                            started_at=run.started_at,
                            cancelled_at=run.cancelled_at,
                            failed_at=run.failed_at,
                            completed_at=run.completed_at,
                            incomplete_details=run.incomplete_details,
                            model=run.model,
                            instructions=run.instructions,
                            tools=run.tools,
                            metadata=run.meta,
                            usage=run.usage,
                            temperature=run.temperature,
                            top_p=run.top_p,
                            max_prompt_tokens=run.max_prompt_tokens,
                            max_completion_tokens=run.max_completion_tokens,
                            truncation_strategy=run.truncation_strategy,
                            tool_choice=run.tool_choice,
                            parallel_tool_calls=run.parallel_tool_calls,
                            response_format=run.response_format,
                        )
                        for run in runs
                    ],
                    first_id=first_id,
                    last_id=last_id,
                    has_more=has_more
                )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")

@router.get("/threads/{thread_id}/runs/{run_id}", response_model=RunObject)
async def retrieve_run(
    thread_id: str,
    run_id: str,
) -> RunObject:
    try:
        thread = db.get_thread_by_id(thread_id)
        if not thread:
            raise HTTPException(status_code=404, detail="Thread not found.")
        
        run = db.get_run_by_id(run_id)
        if not run:
            raise HTTPException(status_code=404, detail="Run not found.")
        
        return RunObject(
            id=run.id,
            created_at=int(run.created_at.timestamp()),
            thread_id=thread_id,
            assistant_id=run.assistant_id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")
    
@router.post("/threads/{thread_id}/runs/{run_id}", response_model=RunObject)
async def modify_run(thread_id: str, run_id: str, request: ModifyRunRequest):
    try:
        run = db.update_run_metadata(thread_id, run_id, request.metadata)
        if not run:
            raise HTTPException(status_code=404, detail="Message not found.")
        
        return  RunObject(
            id=run.id,
            created_at=int(run.created_at.timestamp()),
            thread_id=thread_id,
            assistant_id=run.assistant_id,
            status=run.status,
            required_action=run.required_action,
            last_error=run.last_error,
            expires_at=run.expires_at,
            started_at=run.started_at,
            cancelled_at=run.cancelled_at,
            failed_at=run.failed_at,
            completed_at=run.completed_at,
            incomplete_details=run.incomplete_details,
            model=run.model,
            instructions=run.instructions,
            tools=run.tools,
            metadata=run.meta,
            usage=run.usage,
            temperature=run.temperature,
            top_p=run.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
    except HTTPException as http_exc:
        raise http_exc
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Internal Server Error: {str(e)}")


from fastapi import Request, HTTPException
from starlette.middleware.base import BaseHTTPMiddleware

class APIKeyMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, api_key: str):
        super().__init__(app)
        self.api_key = api_key

    async def dispatch(self, request: Request, call_next):
        auth_header = request.headers.get("Authorization")
        if not auth_header or not auth_header.startswith("Bearer "):
            raise HTTPException(status_code=401, detail="Missing or invalid API key.")

        api_key = auth_header.split("Bearer ")[1]
        if api_key != self.api_key:
            raise HTTPException(status_code=403, detail="Invalid API key.")

        response = await call_next(request)
        return response



from .base_worker import BaseWorker
from .async_worker import AsyncWorker
from .thread_worker import ThreadWorker

class WorkerFactory:
    @staticmethod
    def create_worker(worker_type: str, *args, **kwargs) -> BaseWorker:
        if worker_type == "async":
            return AsyncWorker(*args, **kwargs)
        elif worker_type == "thread":
            return ThreadWorker(*args, **kwargs)
        else:
            raise ValueError(f"Unknown worker type: {worker_type}")


import os
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from llamphouse.core.assistant import Assistant
from .models import Thread, Message, Run, RunStep
from ..types import thread, message, run
from ..types.enum import run_status, run_step_status
from dotenv import load_dotenv
import uuid

load_dotenv()

DATABASE_URL = os.getenv("DATABASE_URL", "postgresql://postgres:password@localhost/llamphouse")

engine = create_engine(DATABASE_URL)

SessionLocal = sessionmaker(autocommit=False, bind=engine)
db = SessionLocal()

def insert_thread(threads: thread.CreateThreadRequest):
    try:
        custom_thread_id = None
        if threads.metadata:
            custom_thread_id = threads.metadata.get("thread_id")
        if custom_thread_id:
            thread_id = custom_thread_id
        else:
            thread_id = str(uuid.uuid4())
        item = Thread(
            id=thread_id,
            name=thread_id,
            tool_resources=threads.tool_resources,
            meta=threads.metadata
        )
        db.add(item)
        db.commit()
        return item
    except Exception as e:
        db.rollback()
        print(f"An error occurred: {e}")
        return None
    
def insert_message(thread_id: str, message: message.CreateMessageRequest):
    try:
        custom_message_id = None
        if message.metadata:
            custom_message_id = message.metadata.get("message_id")
        if custom_message_id:
            message_id = custom_message_id
        else:
            message_id = str(uuid.uuid4())
        item = Message(
            id=message_id,
            role=message.role,
            content=message.content,
            attachments=message.attachments,
            meta=message.metadata or {},
            thread_id=thread_id
        )
        db.add(item)
        db.commit()
        return item
    except Exception as e:
        db.rollback()
        print(f"An error occurred: {e}")
        return None
    
def insert_run(thread_id: str, run: run.RunCreateRequest, assistant: Assistant):
    try:
        custom_run_id = None
        if run.metadata:
            custom_run_id = run.metadata.get("run_id")
        if custom_run_id:
            run_id = custom_run_id
        else:
            run_id = str(uuid.uuid4())
        item = Run(
            id=run_id,
            thread_id=thread_id,
            assistant_id=run.assistant_id,
            model=run.model or assistant.model,
            instructions=run.instructions or assistant.instructions,
            tools=run.tools or assistant.tools,
            meta=run.metadata or {},
            temperature=run.temperature or assistant.temperature,
            top_p=run.top_p or assistant.top_p,
            max_prompt_tokens=run.max_prompt_tokens,
            max_completion_tokens=run.max_completion_tokens,
            truncation_strategy=run.truncation_strategy,
            tool_choice=run.tool_choice,
            parallel_tool_calls=run.parallel_tool_calls,
            response_format=run.response_format,
        )
        db.add(item)
        db.commit()
        return item
    except Exception as e:
        db.rollback()
        print(f"An error occurred: {e}")
        return None

def insert_run_step(run_id: str, assistant_id: str, thread_id: str, step_type: str, step_details: dict, status: str = run_step_status.IN_PROGRESS):
    try:
        run_step_id = str(uuid.uuid4())
        item = RunStep(
            id=run_step_id,
            object="thread.run.step",
            assistant_id=assistant_id,
            thread_id=thread_id,
            run_id=run_id,
            type=step_type,
            status=status,
            step_details=step_details
        )
        db.add(item)
        db.commit()
        db.refresh(item)
        return item
    except Exception as e:
        db.rollback()
        print(f"An error occurred: {e}")
        return None
    
def get_run_by_id(run_id: str):
    try:
        run = db.query(Run).filter(Run.id == run_id).first()
        return run
    except Exception as e:
        print(f"An error occurred: {e}")
        return None

def get_thread_by_id(thread_id: str):
    try:
        thread = db.query(Thread).filter(Thread.id == thread_id).first()
        return thread
    except Exception as e:
        print(f"An error occurred: {e}")
        return None

def get_message_by_id(message_id: str):
    try:
        message = db.query(Message).filter(Message.id == message_id).first()
        return message
    except Exception as e:
        print(f"An error occurred: {e}")
        return None

def get_messages_by_thread_id(
        thread_id: str, 
        limit: int = 20, 
        order: str = "desc", 
        after: str = None, 
        before: str = None
    ) -> list[Message]:
    try:
        query = db.query(Message).filter(Message.thread_id == thread_id)
        if order == "asc":
            query = query.order_by(Message.created_at.asc())
        else:
            query = query.order_by(Message.created_at.desc())
        if after:
            query = query.filter(Message.id > after)
        if before:
            query = query.filter(Message.id < before)
        return query.limit(limit).all()
    except Exception as e:
        print(f"An error occurred: {e}")
        return []
    
def get_runs_by_thread_id(
        thread_id: str, 
        limit: int = 20, 
        order: str = "desc", 
        after: str = None, 
        before: str = None
    ) -> list[Message]:
    try:
        query = db.query(Run).filter(Run.thread_id == thread_id)
        if order == "asc":
            query = query.order_by(Run.created_at.asc())
        else:
            query = query.order_by(Run.created_at.desc())
        if after:
            query = query.filter(Run.id > after)
        if before:
            query = query.filter(Run.id < before)
        return query.limit(limit).all()
    except Exception as e:
        print(f"An error occurred: {e}")
        return []

def get_pending_runs():
    try:
        pending_runs = db.query(Run).filter(Run.status == run_status.QUEUED).all()
        return pending_runs
    except Exception as e:
        print(f"An error occurred while fetching pending runs: {e}")
        return []
    
def get_pending_run():
    try:
        pending_runs = db.query(Run).filter(Run.status == run_status.QUEUED).with_for_update().first()
        return pending_runs
    except Exception as e:
        print(f"An error occurred while fetching pending runs: {e}")
        return None

def list_run_steps(thread_id: str, run_id: str):
    try:
        run_steps = db.query(RunStep).filter(RunStep.run_id == run_id, RunStep.thread_id == thread_id)
        return run_steps
    except Exception as e:
        print(f"An error occurred while fetching run steps: {e}")
        return []

def update_thread_metadata(thread_id: str, metadata: dict):
    try:
        thread = db.query(Thread).filter(Thread.id == thread_id).first()
        if thread:
            thread.meta = metadata
            db.commit()
            return thread
        return None
    except Exception as e:
        db.rollback()
        print(f"An error occurred while updating thread metadata: {e}")
        return None

def update_message_metadata(thread_id: str, message_id: str, metadata: dict):
    try:
        message = db.query(Message).filter(Message.thread_id == thread_id, Message.id == message_id).first()
        
        if message:
            message.meta = metadata or {}
            db.commit()
            return message
        else:
            return None
    except Exception as e:
        db.rollback()
        print(f"An error occurred: {e}")
        return None

def update_run(run: Run):
    try:
        db.merge(run)
        db.commit()
        return True
    except Exception as e:
        db.rollback()
        print(f"An error occurred while updating the run: {e}")
        return False
    
def update_run_metadata(thread_id: str, run_id: str, metadata: dict):
    try:
        run = db.query(Run).filter(Run.thread_id == thread_id, Run.id == run_id).first()
        if run:
            run.meta = metadata
            db.commit()
            return run
        return None
    except Exception as e:
        db.rollback()
        print(f"An error occurred while updating thread metadata: {e}")
        return None
    
def update_run_status(run_id: str, status: str, error: dict = None):
    try:
        run = db.query(Run).filter(Run.id == run_id).first()
        if run:
            run.status = status
            run.last_error = error
            db.commit()
            return run
        return None
    except Exception as e:
        db.rollback()
        print(f"An error occurred while updating the run: {e}")
        return None
    
def update_run_step_status(run_step_id: str, status: str, output = None, error: str = None):
    try:
        run_step = db.query(RunStep).filter(RunStep.id == run_step_id).first()
        if run_step:
            run_step.status = status
            run_step.last_error = error
            if output:
                run_step.step_details["tool_calls"][0]["function"]["output"] = output
            db.commit()
            return run_step
        return None
    except Exception as e:
        db.rollback()
        print(f"An error occurred while updating the run step: {e}")
        return None

def delete_thread_by_id(thread_id: str):
    try:
        thread = db.query(Thread).filter(Thread.id == thread_id).first()
        if thread:
            db.delete(thread)
            db.commit()
            return True
        return False
    except Exception as e:
        db.rollback()
        print(f"An error occurred: {e}")
        return False

def delete_messages_by_thread_id(thread_id: str):
    try:
        db.query(Message).filter(Message.thread_id == thread_id).delete()
        db.commit()
        return True
    except Exception as e:
        db.rollback()
        print(f"An error occurred: {e}")
        return False

def delete_message_by_id(message_id: str):
    try:
        message = db.query(Message).filter(Message.id == message_id).first()
        if message:
            db.delete(message)
            db.commit()
            return True
        return False
    except Exception as e:
        db.rollback()
        print(f"An error occurred: {e}")
        return False

def delete_message(thread_id: str, message_id: str):
    try:
        message = db.query(Message).filter(
            Message.thread_id == thread_id, Message.id == message_id).first()
        
        if message:
            db.delete(message)
            db.commit()
            return True
        else:
            return False
    except Exception as e:
        db.rollback()
        print(f"An error occurred: {e}")
        return False


import sys
from typing import List, Optional
import uvicorn
from fastapi import FastAPI
from .routes import all_routes
from .assistant import Assistant
from fastapi.middleware.cors import CORSMiddleware
from llamphouse.core.middlewares.api_key_middleware import APIKeyMiddleware
import asyncio
from .workers.factory import WorkerFactory

class LLAMPHouse:
    def __init__(self, assistants: List[Assistant] = [], worker_type = "thread", api_key: Optional[str] = None):
        self.assistants = assistants
        self.worker_type = worker_type
        self.api_key = api_key
        self.fastapi = FastAPI(title="LLAMPHouse API Server")
        self.fastapi.state.assistants = assistants
        self.fastapi.state.task_queues = {}
        if self.api_key:
            self.fastapi.add_middleware(APIKeyMiddleware, api_key=self.api_key)
        self._register_routes()

    def __print_ignite(self, host, port):
        ascii_art = """
                  __,--'
       .-.  __,--'
      |  o| 
     [IIIII]`--.__
      |===|       `--.__
      |===|
      |===|
      |===|
______[===]______
"""
        print(ascii_art)
        print("We have light!")
        print(f"LLAMPHOUSE server running on http://{host}:{port}")
        sys.stdout.flush()

    def ignite(self, host="0.0.0.0", port=80, reload=False):
        # Start workers based on the specified type
        @self.fastapi.on_event("startup")
        async def startup_event():
            loop = asyncio.get_running_loop()
            # create a worker and start it
            worker = WorkerFactory.create_worker(self.worker_type, self.assistants, self.fastapi.state, loop)
            worker.start()

        self.__print_ignite(host, port)
        uvicorn.run(self.fastapi, host=host, port=port, reload=reload)

    def _register_routes(self):       
        for router in all_routes:
            self.fastapi.include_router(router)



from llamphouse.core import LLAMPHouse, Assistant
from typing import List, Optional
from dotenv import load_dotenv
from llamphouse.core.context import Context

load_dotenv(override=True)

class CustomAssistant(Assistant):
    def __init__(
        self, 
        name: str, 
        model: str, 
        temperature: float = 0.7, 
        top_p: float = 1.0, 
        instructions: Optional[str] = None,
        id: Optional[str] = None, 
        description: Optional[str] = None, 
        tools: Optional[List[str]] = None
    ):
        super().__init__(
            name=name, 
            model=model, 
            temperature=temperature, 
            top_p=top_p,
            instructions=instructions,
            id=id, 
            description=description, 
            tools=tools
        )

    def run(self, context: Context, *args, **kwargs):
        print(f"thread: {context.thread_id} ")
        print(f"assistant: {context.assistant_id} ")

assistant1 = CustomAssistant(
    id="Chatbot1",
    name="ChatBot", 
    model="gpt-3.5-turbo", 
    description="A general-purpose chatbot", 
    temperature=0.8, 
    instructions="You are a personal math tutor. When asked a question, write and run Python code to answer the question.",
    top_p=0.9, 
    # tools=["text-generation"]
)
assistant2 = CustomAssistant(
    id="HelpBot1",
    name="HelpBot", 
    model="gpt-4", 
    description="A help desk assistant", 
    temperature=0.7, 
    instructions="You are a personal math tutor. When asked a question, write and run Python code to answer the question.",
    top_p=1.0, 
    # tools=["text-generation", "code_interpreter"]
)
assistant3 = CustomAssistant(
    id="CustomerBot1",
    name="CustomerBot", 
    model="gpt-4", 
    description="A help desk assistant", 
    temperature=0.7, 
    instructions="You are a personal math tutor. When asked a question, write and run Python code to answer the question.",
    top_p=1.0, 
    # tools=["text-generation", "code_interpreter"]
)

def main():
    llamphouse = LLAMPHouse(assistants=[assistant1, assistant2, assistant3], api_key="123")
    llamphouse.ignite(host="127.0.0.1", port=8000)

if __name__ == "__main__":
    main()



from openai import OpenAI

def main():
    client = OpenAI(api_key="123", base_url="http://127.0.0.1:8000")
    
    # List assistant in glowstick
    my_assistants = client.beta.assistants.list()
    # print(my_assistants)
    # # Create new thread and message in db
    empty_thread = client.beta.threads.create()
    print(f"Created thread: {empty_thread.id}")
    new_message = client.beta.threads.messages.create(empty_thread.id, role="user", content="How are you")
    print(f"Added message: {new_message.id}")

    # Create new run in specific thread & assistant
    print("Starting new run")
    new_run =client.beta.threads.runs.create_and_poll(assistant_id="Chatbot1", thread_id=empty_thread.id, temperature=1)
    print(f"Run finished: {new_run}")

if __name__ == "__main__":
    main()
